// PL0.h 

#include <stdbool.h>
#include <stdio.h>

#define norw		11		// no. of reserved words
#define txmax		100		// length of identifier table
#define nmax		9		// max. no. of digits in numbers
#define al		10		// length of identifiers
#define amax		2047		// maximum of constant value
#define levmax		3		// maximum depth of block nesting
#define cxmax		2000		// size of code array
#define dxmax		500		// maximum of variable address
#define errmax		300		// maximum of errors

#define nul		0x1		// 1
#define ident		0x2		// 2
#define number		0x4		// 4
#define plus		0x8		// 8
#define minus		0x10		// 16
#define times		0x20		// 32
#define slash		0x40		// 64
#define oddsym		0x80		// 128
#define eql		0x100		// 256
#define neq		0x200		// 512
#define lss		0x400		// 1024
#define leq		0x800		// 2048
#define gtr		0x1000		// 4096
#define geq		0x2000		// 8192
#define lparen		0x4000		// 16384
#define rparen		0x8000		// 32768
#define comma		0x10000		// 65536
#define semicolon	0x20000		// 131072
#define period		0x40000		// 262144
#define becomes		0x80000		// 524288
#define beginsym	0x100000	// 1048576
#define endsym		0x200000	// 2097152
#define ifsym		0x400000	// 4194304
#define thensym		0x800000	// 8388608
#define whilesym	0x1000000	// 16777216
#define dosym		0x2000000	// 33554432
#define callsym		0x4000000	// 67108864
#define constsym	0x8000000	// 134217728
#define varsym		0x10000000	// 268435456
#define procsym		0x20000000	// 536870912

enum object {
    constant, variable, proc
};

enum fct {
    lit, opr, lod, sto, cal, Int, jmp, jpc	// functions
};

typedef struct{
    enum fct f;		// function code
    long l;		// level
    long a;		// displacement address
} instruction;
/*  lit 0, a : load constant a
    opr 0, a : execute operation a
    lod l, a : load variable l, a
    sto l, a : store variable l, a
    cal l, a : call procedure a at level l
    Int 0, a : increment t-register by a
    jmp 0, a : jump to a
    jpc 0, a : jump conditional to a       */

extern long sym;			// last symbol read
extern char id[al+1];			// last identifier read
extern long num;			// last number read

extern long ll;			// line length
extern long cc;			// character count

extern char line[81];		// 下标为0的单元不使用。
//@ weak global invariant last_char_in_line : line[ll] == ' ';
extern char ch;			// last character read

extern long kk;
//@ strong global invariant ident_length : 0 <= kk <= al;
extern long err;
//@ strong global invariant error_number : 0 <= err <= errmax;


extern char a[al+1];
//@ weak global invariant last_char_in_a : \is_pstring(a, al) && (\forall long m:[kk..al-1]. a[m] == ' ');

extern instruction code[cxmax+1];
extern long cx;			// code allocation index
/*@ strong global invariant cx_inv: 0 <= cx <= cxmax+1; */

extern char word[norw][al+1];
/*@ predicate reservedWord(char word[norw][al+1]) =
	(\forall long i:[0..norw-1].\is_pstring(word[i], al)) &&
	\string(word[0], al) == "begin     " && \string(word[1], al) == "call      " &&
	\string(word[2], al) == "const     " && \string(word[3], al) == "do        " &&
	\string(word[4], al) == "end       " && \string(word[5], al) == "if        " &&
	\string(word[6], al) == "odd       " && \string(word[7], al) == "procedure " &&
	\string(word[8], al) == "then      " && \string(word[9], al) == "var       " &&
	\string(word[10], al) == "while     " &&
	(\forall long i:[0..9]. \string(word[i], al) < \string(word[i+1], al));
    strong global invariant reserved_word_table:
	 reservedWord(word);
*/
extern long wsym[norw];
/*@ predicate reservedWordToken(unsigned long wsym[norw]) =
	wsym[0] == beginsym &&
	wsym[1] == callsym && wsym[2] == constsym && wsym[3] == dosym &&
	wsym[4] == endsym && wsym[5] == ifsym && wsym[6] == oddsym &&
	wsym[7] == procsym && wsym[8] == thensym && wsym[9] == varsym &&
	wsym[10] == whilesym;
    strong global invariant reserved_word_token:
	reservedWordToken(wsym);
*/

extern long ssym[256]; 
/*@ predicate operatorToken(unsigned long ssym[256]) =
	(\forall long i:[0..39]. ssym[i] == nul )&&
	ssym[40] == lparen && ssym[41] == rparen && ssym[42] == times && ssym[43] == plus &&
	ssym[44] == comma && ssym[45] == minus && ssym[46] == period && ssym[47] == slash &&
	(\forall long i:[48..58]. ssym[i] == nul) && ssym[59] == slash && ssym[60] == nul &&
	ssym[61] == eql && (\forall long i:[62..255]. ssym[i] == nul);
    strong global invariant reserved_word_token:
	operatorToken(ssym);
*/

extern char mnemonic[8][3+1];
/*@ strong global invariant mnemonic_table:
    (\forall long i:[0..8-1].\is_pstring(mnemonic[i], 3)) &&
    \string(mnemonic[0], 3) == "lit" && \string(mnemonic[1], 3) == "opr" &&
    \string(mnemonic[2], 3) == "lod" && \string(mnemonic[3], 3) == "sto" && \string(mnemonic[4], 3) == "cal" &&
    \string(mnemonic[5], 3) == "int" && \string(mnemonic[6], 3) == "jmp" && \string(mnemonic[7], 3) == "jpc";
*/

extern long declbegsys;
//@ strong global invariant declbegsys_inv: declbegsys == (constsym | varsym | procsym);

extern long statbegsys;
//@ strong global invariant statbegsys_inv: statbegsys == (beginsym | callsym | ifsym | whilesym);
    
extern long facbegsys;
//@ strong global invariant facbegsys_inv: facbegsys == (ident | number | lparen);



extern long tx;		// current table index
//@ strong global invariant tx_inv: 0 <= tx <= txmax;

struct name_tab {
    char name[al+1];
    enum object kind;
    long val;
    long level;
    long addr;
};
extern struct name_tab table[txmax+1];	//对于原来的代码来说，下面的强变量不变式并不妥当。为了维持这个不变式，在miscellany.c文件的enter函数中增加了一个对level域的赋值语句。
//@ strong global invariant table_inv: \forall long i:[1..tx]. (\is_pstring(table[i].name, al) && 0 <= table[i].level <= levmax);

extern char infilename[80];
extern FILE* infile;

// the following variables for block
extern long dx;			// data allocation index
//@ strong global invariant dx_inv: 0 <= dx <= dxmax+1;
extern long lev;		// current depth of block nesting
//@ strong global invariant level_inv:  0 <= lev <= levmax;

// the following array space for interpreter
#define stacksize 50000
extern long s[stacksize];	// datastore

